#include "Plane.h"
#include "ImplicitOctTree.h"
#include "LU.h"

Plane::Plane(PointSet* ps, float R, ImplicitOctCell *cell){
		int* index_list;
			
		float *c = cell->_center;

		float minB[] = {c[0] - R, c[1] - R, c[2] - R};
		float maxB[] = {c[0] + R, c[1] + R, c[2] + R};

		int listN = ps->_tree->collectPointsInBox(index_list, minB, maxB);
		cell->_pointN = listN;

		if(listN == 0)
			return;
		
		float (*point)[3] = ps->_point;
		float (*normal)[3] = ps->_normal;

		double p_ave[] = {0,0,0};
		double n_ave[] = {0,0,0};

		//For LU
		float** A = new float*[5];
		A[1] = new float[5];
		A[2] = new float[5];
		A[3] = new float[5];
		A[4] = new float[5];
		A[1][1] = A[1][2] = A[1][3] = A[1][4] = 
		A[2][1] = A[2][2] = A[2][3] = A[2][4] =
		A[3][1] = A[3][2] = A[3][3] = A[3][4] = 
		A[4][1] = A[4][2] = A[4][3] = A[4][4] = 0;
		int* index = new int[5];
		float d;
		double* b = new double[5];
		b[1] = b[2] = b[3] = b[4] = 0;

		float off = R*OFF;
		int count = 0;
		for(int i=0; i<listN; i++){
			int j = index_list[i];
			float* p = point[j];
			float* n = normal[j];

			float x = p[0] + off*n[0];
			float y = p[1] + off*n[1];
			float z = p[2] + off*n[2];

			float vx = x - c[0];
			float vy = y - c[1];
			float vz = z - c[2];

			double w = cell->weight(sqrt(vx*vx+vy*vy+vz*vz), R);
			if(w != 0){
				count++;
				w *= w;

				A[1][1] += w;
				A[1][2] += w*x;
				A[1][3] += w*y;
				A[1][4] += w*z;

				A[2][1] += w*x;
				A[2][2] += w*x*x;
				A[2][3] += w*y*x;
				A[2][4] += w*z*x;
			
				A[3][1] += w*y;
				A[3][2] += w*x*y;
				A[3][3] += w*y*y;
				A[3][4] += w*z*y;

				A[4][1] += w*z;
				A[4][2] += w*x*z;
				A[4][3] += w*y*z;
				A[4][4] += w*z*z;
				
				b[1] += w*off;
				b[2] += w*x*off;
				b[3] += w*y*off;
				b[4] += w*z*off;
			}

			x = p[0] - off*n[0];
			y = p[1] - off*n[1];
			z = p[2] - off*n[2];

			vx = p[0] - c[0];
			vy = p[1] - c[1];
			vz = p[2] - c[2];

			w = cell->weight(sqrt(vx*vx+vy*vy+vz*vz), R);
			if(w != 0){
				count++;
				w *= w;

				A[1][1] += w;
				A[1][2] += w*x;
				A[1][3] += w*y;
				A[1][4] += w*z;

				A[2][1] += w*x;
				A[2][2] += w*x*x;
				A[2][3] += w*y*x;
				A[2][4] += w*z*x;
			
				A[3][1] += w*y;
				A[3][2] += w*x*y;
				A[3][3] += w*y*y;
				A[3][4] += w*z*y;

				A[4][1] += w*z;
				A[4][2] += w*x*z;
				A[4][3] += w*y*z;
				A[4][4] += w*z*z;
				
				b[1] -= w*off;
				b[2] -= w*x*off;
				b[3] -= w*y*off;
				b[4] -= w*z*off;
			}
		}
		if(count < 4){
			delete[] A[1];
			delete[] A[2];
			delete[] A[3];
			delete[] A[4];
			delete[] A;
			delete[] index;
			delete[] b;
			cell->_pointN = 0;
			return;
		}

		LU lu;
		lu.ludcmp(A, 4, index, &d);
		lu.lubksb(A, 4, index, b);

		_c0 = b[1];
		_cx = b[2];
		_cy = b[3];
		_cz = b[4];

		delete[] A[1];
		delete[] A[2];
		delete[] A[3];
		delete[] A[4];
		delete[] A;
		delete[] index;
		delete[] b;

		//normalize function
		double len = sqrt(_cx*_cx + _cy*_cy + _cz*_cz);
		if(len != 0){
			_cx = (float)(_cx/len);
			_cy = (float)(_cy/len);
			_cz = (float)(_cz/len);
			_c0 = (float)(_c0/len);
		}
		else{
			cell->_pointN = 0;
			return;
		}

		//max error
		_error = 0;
		for(i=0; i<listN; i++){
			int j = index_list[i];
			float* p = point[j];

			float f = fabs(value(p[0], p[1], p[2]));
			if(f > _error)
				_error = f;
		}
}